"""
	This module contains the interface for evaluators that can assess the performance of the variants of the 'SeqClu'
	algorithm.
"""
from abc import ABC, abstractmethod


class IEvaluator(ABC):

	# PUBLIC METHODS #

	@abstractmethod
	def evaluate(self) -> None:
		"""
			This method evaluates the performance of the variants of the 'SeqClu' algorithm.

			:return: void
		"""
		pass



"""
    This package contains utility methods that are used throughout the 'SeqClu' framework.
"""

from .calculate_f1_score import calculate_f1_score
from .construct_stream import constructStream
from .hash_filter import isHashInKey
from .hash_sequence import hashSequence
from .statistical_testing import statistical_test

__all__ = ["calculate_f1_score",
           "constructStream",
           "isHashInKey",
           "hashSequence",
           "statistical_test"]



"""
    This module contains a naive implementation of the online version of the 'SeqClu' algorithm that is used to compare
    the results of the developed variants of the 'SeqClu' algorithm to the results of the implementation contained in
    this module that serves as the baseline.
"""

from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from seqclupv.library.data_source.fake_data_source import FakeDataSource
from seqclupv.library.interfaces.distance_measure import IDistanceMeasure
from seqclupv.library.interfaces.seqclu import ISeqClu
from seqclupv.library.utilities.hash_sequence import hashSequence


class SeqCluBaselineOnline(ISeqClu):

    # PROPERTIES #

    @property
    def clusters(self) -> ndarray:
        return self._clusters

    @property
    def distances(self) -> Dict[Tuple[str, str], float]:
        return self._distances

    @property
    def labels(self) -> Dict[str, Union[chr, int]]:
        return self._labels

    @property
    def finalLabels(self) -> Dict[str, Union[chr, int]]:
        """
            This property stores the correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.

            :return: The correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.
        """
        result = {}
        for clusterIdx, prototypeHashes in enumerate(self.clusters):
            for prototypeHash in prototypeHashes:
                result[prototypeHash] = self.classes[clusterIdx]
        result.update(self.labels)
        return result

    @property
    def numProcessed(self) -> int:
        return self._numProcessed

    @property
    def prototypes(self) -> List[Dict[str, ndarray]]:
        return self._prototypes

    # CONSTRUCTOR #

    def __init__(self, dataSource: FakeDataSource, distanceMeasure: IDistanceMeasure,
                 numClusters: int, numPrototypes: int) -> None:
        """
            This method initializes the 'SeqClu' algorithm with a data source, distance measure, number of clusters
            and number of prototypes.

            :param dataSource: The data source that should be used in the 'SeqClu' algorithm.
            :param distanceMeasure: The distance measure that should be used in the 'SeqClu' algorithm.
            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.
            :return: void
        """
        super().__init__(dataSource, distanceMeasure, numClusters, numPrototypes)
        self._distances = {}
        self._labels = {}
        self._numProcessed = 0
        prototypesList = []
        for _ in range(numClusters):
            prototypesList.append({})
        self._prototypes = prototypesList
        self._clusters = self._initializeClusters(dataSource)

    # PUBLIC METHODS #

    def alreadyProcessed(self, sequenceHash: str) -> bool:
        # Check if the sequence is already stored as a prototype of any of the clusters or as a candidate.
        if sequenceHash in self.labels:
            return True
        for clusterIdx in range(self.numClusters):
            if sequenceHash in self.prototypes[clusterIdx]:
                return True
        return False

    def execute(self) -> None:
        """
            This method executes the 'SeqClu' algorithm from start to finish. The main loop over the data is
            implemented in here.

            :return: void
        """
        iteration = 0
        while not self.finish:
            iteration += 1
            print(f"[SeqClu] Iteration {iteration}.")
            finish = self._advanceTick()
            if finish:
                print("[SeqClu] The algorithm finished executing. Processing the buffer and returning...")
                break

    def processSequence(self, sequence: Tuple[str, ndarray], considerCandidacy: bool) -> None:
        """
            This method processes one sequence of the data set that is processed in the 'SeqClu' algorithm.

            :param sequence: The sequence that should be processed in the 'SeqClu' algorithm.
            :param considerCandidacy: A boolean value indicating whether or not the algorithm should consider if the
            sequence could be a prototype for any of the clusters.
            :return: void
        """
        if self.alreadyProcessed(sequence[0]):
            return
        sequenceHash, sequenceData = sequence
        minDistance = np.inf
        minIdx = -1

        for clusterIdx, cluster in enumerate(self.clusters):
            distance = 0
            for prototypeHash in cluster:
                distance += self._getDistance(clusterIdx, sequence, (prototypeHash, None))
            distance = distance / self.numPrototypes

            if minDistance > distance:
                minDistance = distance
                minIdx = clusterIdx

        self._updatePrototypes(minIdx, sequence)

        self._labels[sequenceHash] = self.classes[minIdx]

    # PRIVATE METHODS #

    def isPrototypeForMultiple(self, prototypeHash: str) -> bool:
        counter = 0
        for clusterIdx in range(self.numClusters):
            counter += np.count_nonzero(self.clusters[clusterIdx] == prototypeHash)
        if counter == 0:
            raise ValueError("Invalid function call, the provided hash does not correspond to a prototype for any"
                             " of the clusters.")
        return counter > 1

    def _getDistance(self, clusterIdx: int, sequenceOne: Tuple[str, Optional[ndarray]],
                     sequenceTwo: Tuple[str, Optional[ndarray]]) -> float:
        sequenceHashOne, sequenceDataOne = sequenceOne
        sequenceHashTwo, sequenceDataTwo = sequenceTwo

        if (sequenceHashOne, sequenceHashTwo) in self.distances:
            return self.distances[(sequenceHashOne, sequenceHashTwo)]
        if (sequenceHashTwo, sequenceHashOne) in self.distances:
            return self.distances[(sequenceHashTwo, sequenceHashOne)]
        if sequenceDataOne is None:
            sequenceDataOne = self.prototypes[clusterIdx][sequenceHashOne]
        if sequenceDataTwo is None:
            sequenceDataTwo = self.prototypes[clusterIdx][sequenceHashTwo]

        distance = self.distanceMeasure.calculateDistance(sequenceDataOne, sequenceDataTwo)
        self.distances[(sequenceHashOne, sequenceHashTwo)] = distance
        self.distances[(sequenceHashTwo, sequenceHashOne)] = distance

        return distance

    def _initializeClusters(self, dataSource: FakeDataSource) -> ndarray:
        result = np.empty((self.numClusters, self.numPrototypes), dtype=object)
        # NOTE: The first numClusters * numPrototypes sequences could contain identical sequences.
        for i in range(self.numClusters):
            for j, sequence in enumerate(dataSource.data[i * self.numPrototypes:i * self.numPrototypes + self.numPrototypes]):
                sequenceHash = hashSequence(sequence)
                result[i][j] = sequenceHash
                self.prototypes[i][sequenceHash] = sequence
        dataSource.currentIndex = self.numClusters * self.numPrototypes
        return result

    def _updatePrototypes(self, minIdx: int, sequence: Tuple[str, ndarray]) -> None:
        sequenceHash, sequenceData = sequence
        prototypeHashes = self.clusters[minIdx]
        sums = []
        for i, prototypeHashOne in enumerate(prototypeHashes):
            s = 0
            for j, prototypeHashTwo in enumerate(prototypeHashes):
                if i != j:
                    s += self._getDistance(minIdx, (prototypeHashOne, None), (prototypeHashTwo, None))
            sums.append(s)
        zipped = list(zip(sums, [i for i in range(self.numPrototypes)]))
        zipped.sort(key=lambda x: x[0])

        toReplaceIdx = zipped[-1][1]
        toReplaceHash = prototypeHashes[toReplaceIdx]
        assert self.clusters[minIdx][toReplaceIdx] == toReplaceHash
        if not self.isPrototypeForMultiple(toReplaceHash):
            del self.prototypes[minIdx][toReplaceHash]
        self.clusters[minIdx][toReplaceIdx] = sequenceHash
        self.labels[toReplaceHash] = self.classes[minIdx]
        self.prototypes[minIdx][sequenceHash] = sequenceData



"""
    This module contains an implementation of a data structure that stores how often a prototype has been observed to
    be the closest to one of the prototypes of that cluster.
"""

from typing import Dict, Optional, Set, Tuple

import numpy as np


class PrototypeFrequencyStore:

    # STATIC METHODS #

    @staticmethod
    def computeSumOfDistances(prototypeHashes: Set[str], distances: Dict[Tuple[str, str], float]) -> float:
        """
            This method computes the sum of distances between all pairs of prototypes identified by a set of prototype
            hashes given a dictionary of pair-wise distances.

            :param prototypeHashes: The hashes of the prototypes for which
            the sum of pair-wise distances needs to be computed.
            :param distances: The pair-wise distances stored in a dictionary identified by a tuple of prototype hashes
            that identify the two prototypes.
            :return: The sum of distances for the given prototypes and dictionary of distances.
        """
        result = 0
        for prototypeOneHash in prototypeHashes:
            for prototypeTwoHash in prototypeHashes:
                if prototypeOneHash == prototypeTwoHash:
                    continue
                if (prototypeOneHash, prototypeTwoHash) in distances:
                    result += distances[(prototypeOneHash, prototypeTwoHash)]
                elif (prototypeTwoHash, prototypeOneHash) in distances:
                    result += distances[(prototypeTwoHash, prototypeOneHash)]
                else:
                    raise ValueError("[SeqClu] Distance is not present in dictionary.")
        return result

    # PROPERTIES #

    @property
    def frequencies(self) -> Dict[str, Optional[int]]:
        """
            This property stores the amount of times a prototype has been observed as being closest to an incoming
            sequence as a dictionary where the keys are the hashes of the prototypes.

            :return: A dictionary where the keys are the hashes of the prototypes and the values are the amounts of
            times a prototype has been observed as being closest to an incoming sequence.
        """
        return self._frequencies

    @property
    def numPrototypes(self) -> int:
        """
            This property stores the number of prototypes that will be used in the 'SeqClu' algorithm.

            :return: An integer representing the number of prototypes that will be used in the 'SeqClu' algorithm.
        """
        return self._numPrototypes

    @property
    def totalObservations(self) -> int:
        """
            This property stores the total number of observations that was made.

            :return: An integer representing the total number of observations that was made.
        """
        return self._totalObservations

    # CONSTRUCTOR #

    def __init__(self, numPrototypes: int) -> None:
        """
            This method initializes the data structure storing information about the prototype frequencies.

            :param numPrototypes: The number of prototypes that will be used in the 'SeqClu' algorithm.
        """
        self._frequencies = {}
        self._numPrototypes = numPrototypes
        self._totalObservations = 0

    # PUBLIC METHODS #

    def closestPrototypeObserved(self, prototypeHash: str, numVotes: int) -> None:
        """
            This method adds a given amount of votes to the amount of votes that a prototype already has.

            :param prototypeHash: The hash of the prototype that should get votes.
            :param numVotes: The number of votes that the prototype should get.
            :return: void
        """
        assert prototypeHash in self.frequencies
        if self.frequencies[prototypeHash] is None:
            self.frequencies[prototypeHash] = numVotes
        else:
            self.frequencies[prototypeHash] += numVotes
        self._totalObservations += numVotes

    def getWeight(self, prototypeHash: str) -> float:
        """
            This method calculates the weight of a prototype, identified by its hash, and returns this value.

            :param prototypeHash: The hash of the prototype for which the weight needs to be calculated and returned.
            :return: The weight of the prototype identified by its hash.
        """
        assert prototypeHash in self.frequencies
        assert len(self.frequencies) == self.numPrototypes
        if self.frequencies[prototypeHash] is None:
            return 0
        result = self.frequencies[prototypeHash] / self.totalObservations
        return self.frequencies[prototypeHash] / self.totalObservations

    def initializePrototype(self, prototypeHash: str) -> None:
        """
            This method initializes a prototype given its hash.

            :param prototypeHash: The hash of the prototype for which an entry needs to be added to the prototype
            frequencies data structure.
            :return: void
        """
        assert prototypeHash not in self.frequencies
        self.frequencies[prototypeHash] = None

    def updatePrototypes(self, newPrototypeHashes: Set[str], addedPrototypeHashes: Set[str],
                         removedPrototypeHashes: Set[str], distances: Dict[Tuple[str, str], float]) -> None:
        """
            This method updates the prototypes that are stored in the prototype frequencies store given sets of
            updated prototype hashes, newly added prototype hashes and rmeoved prototype hashes, as well as
            a dictionary of pair-wise distances for all pairs of prototypes, regardless of whether or not these
            prototypes are discarded.

            :param newPrototypeHashes: The hashes of the updated prototypes.
            :param addedPrototypeHashes: The hashes of the newly dded prototypes.
            :param removedPrototypeHashes: The hashes of the removed prototypes.
            :param distances: The pair-wise distances for all pairs of prototypes, regardless of whether or not these
            prototypes are discarded.
            :return: void
        """
        print(f"[SeqClu] Updating prototypes...")

        # Initialize new prototypes with empty.
        for prototypeHash in addedPrototypeHashes:
            self.frequencies[prototypeHash] = None

        self._distributeVotes(newPrototypeHashes, removedPrototypeHashes, distances)

        # Sanity check.
        assert len(self.frequencies) == self.numPrototypes

    # PRIVATE METHODS #

    def _distributeVotes(self, newPrototypeHashes: Set[str], removedPrototypeHashes: Set[str],
                         distances: Dict[Tuple[str, str], float]) -> None:
        """
            This method distributes the votes of a discarded prototype to all other prototypes in the store.

            :param newPrototypeHashes: The hashes of the updated prototypes.
            :param removedPrototypeHashes: The hashes of the removed prototypes.
            :param distances: The pair-wise distances for all pairs of prototypes, regardless of whether or not these
            prototypes are discarded.
            :return: void
        """
        # 1. Compute sum of distances from prototype to all other prototypes.
        sumOfDistances = PrototypeFrequencyStore.computeSumOfDistances(newPrototypeHashes, distances)

        for toRemovePrototypeHash in removedPrototypeHashes:
            numVotes = self._removePrototype(toRemovePrototypeHash)
            if numVotes == 0:
                continue
            fractions = np.empty((self.numPrototypes,), dtype=float)
            for i, prototypeHash in enumerate(newPrototypeHashes):
                # 2. Compute fraction of distance to some prototype to removed prototype over the sum of distances.
                # This number is the percentage of the distance that comes from that prototype. The lower, the better.
                if (toRemovePrototypeHash, prototypeHash) in distances:
                    fractions[i] = distances[(toRemovePrototypeHash, prototypeHash)] / sumOfDistances
                elif (prototypeHash, toRemovePrototypeHash) in distances:
                    fractions[i] = distances[(prototypeHash, toRemovePrototypeHash)] / sumOfDistances
                else:
                    raise ValueError("Distance is not in dictionary.")
            # 3. Compute 1 - the number in step 2.
            fractions = 1 - fractions
            # 4. Normalize all values by the sum of these values to make sure
            # its sum sums to 1 and the transferred votes are therefore the same.
            fractions /= np.sum(fractions)
            for i, prototypeHash in enumerate(newPrototypeHashes):
                fraction = fractions[i]
                self.closestPrototypeObserved(prototypeHash, int(fraction * numVotes))

    def _removePrototype(self, toRemoveHash: str) -> int:
        """
            This method removes a prototype from the prototype frequencies store.

            :param toRemoveHash: The hash of the prototype that needs to be removed.
            :return: The amount of votes that the prototype had.
        """
        assert toRemoveHash in self.frequencies
        numVotes = self.frequencies[toRemoveHash]
        del self.frequencies[toRemoveHash]
        if numVotes is None:
            return 0
        self._totalObservations -= numVotes
        return numVotes



"""
    This module contains an implementation of a heuristic functions that are used in variants of the 'SeqClu' algorithm.
    The heuristic function calculates the value of a (candidate) prototype based on a linear combination of the
    representativeness and the weight of the (candidate) prototype.
"""

from seqclupv.library.interfaces.heuristics.prototype_value import IPrototypeValue


class LinearPrototypeValue(IPrototypeValue):

    # PROPERTIES #

    @property
    def ratio(self) -> float:
        """
            Returns the value 'a' in a:1 where a:1 is the ratio between the representativeness and the weight.

            :return: The value 'a' in a:1 where a:1 is the ratio between the representativeness and the weight.
        """
        return self._ratio

    # CONSTRUCTORS #

    def __init__(self, ratio: float) -> None:
        """
            This method initializes the function that returns the linear combination of the representativeness and
            weight of a prototype.

            :param ratio: The value 'a' in a:1 where a:1 is the ratio between the representativeness and the weight.
        """
        self._ratio = ratio

    def evaluate(self, representativeness: float, weight: float) -> float:
        """
            This method evaluates the function that returns the linear combination of the representativeness and
            weight of some prototype.

            :param representativeness: The representativeness of some prototype.
            :param weight: The weight of some prototype.
            :return: The value of the prototype.
        """
        weightRatio = 1 / (self.ratio + 1)
        representativenessRatio = 1 - weightRatio
        result = representativenessRatio * representativeness + weightRatio * weight
        return result



"""
    This package contains three data sets, namely:
    1. GesturePebbleZ1 (http://www.timeseriesclassification.com/description.php?Dataset=GesturePebbleZ1)
    2. UJI Pen Characters (https://archive.ics.uci.edu/ml/datasets/UJI+Pen+Characters)
    3. PLAID (http://www.timeseriesclassification.com/description.php?Dataset=PLAID)

    Additionally, the package contains the original source code provided at the start of the project as a Jupyter
    Notebook.
"""

__all__ = []



"""
    This package contains all source code that is used in the variants of the 'SeqClu' algorithm that are contained in
    this framework.
"""

from .data_generator import HandwrittenCharacterGenerator, GesturePebbleGenerator, PLAIDGenerator, CurveGenerator
from .data_source import FakeDataSource
from .distance import DynamicTimeWarping
from .evaluation import BasicBaselinePrototypes, BasicEvaluator
from .heuristics import LinearPrototypeValue
from .interfaces import IBaselinePrototypes, ICandidateStore, IDataGenerator, IDataSource, IDistanceMeasure, \
    IEvaluator, IFakeDataSource, IPrototypeValue, ISeqClu
from .storage import CandidateStore, ClusterStore, PrototypeFrequencyStore, PrototypeStore, SequenceStore
from .utilities import calculate_f1_score, constructStream, isHashInKey, hashSequence, statistical_test
from .visualization import Visualizer


__all__ = ["HandwrittenCharacterGenerator",
           "GesturePebbleGenerator",
           "PLAIDGenerator",
           "CurveGenerator",
           "FakeDataSource",
           "DynamicTimeWarping",
           "BasicBaselinePrototypes",
           "BasicEvaluator",
           "LinearPrototypeValue",
           "IBaselinePrototypes",
           "ICandidateStore",
           "IDataGenerator",
           "IDataSource",
           "IDistanceMeasure",
           "IEvaluator",
           "IFakeDataSource",
           "IPrototypeValue",
           "ISeqClu",
           "CandidateStore",
           "ClusterStore",
           "PrototypeFrequencyStore",
           "PrototypeStore",
           "SequenceStore",
           "calculate_f1_score",
           "constructStream",
           "isHashInKey",
           "hashSequence",
           "statistical_test",
           "Visualizer"]



"""
    This module contains an interface for the data structure that stores the candidate prototypes for all clusters
    that are maintained as part of the 'SeqClu' algorithm.
"""

from abc import ABC, abstractmethod
from typing import Dict, Optional, Set, Tuple

from numpy import ndarray


class ICandidateStore(ABC):

    @property
    @abstractmethod
    def candidates(self) -> Dict[str, Tuple[ndarray, Set[int]]]:
        """
            This property returns the candidate prototypes along with the indices of the clusters that the
            sequences are candidates for that are stored in the candidate store.

            :return: The candidate prototypes stored in the candidate store.
        """
        pass

    @property
    @abstractmethod
    def candidateHistory(self) -> Dict[str, int]:
        """
            This property returns a dictionary where the keys are the candidate hashes and the values are the ticks at
            which the candidate prototypes were last updated.

            :return: A dictionary where the keys are the candidate hashes and the values are the ticks at
            which the candidate prototypes were last updated.
        """
        pass

    @property
    @abstractmethod
    def lastUpdate(self) -> int:
        """
            This property returns the tick at which the candidate store was last updated.

            :return: An integer representing the tick at which the candidate store was last updated.
        """
        pass

    @abstractmethod
    def addToCandidates(self, candidate: ndarray, candidateFor: Set[int], tick: int) -> None:
        """
            This method adds a sequence to the candidate store for a set of clusters at a given tick.

            :param candidate: The sequence that is added as a candidate to the candidate store.
            :param candidateFor: The set of clusters that the sequence is a candidate for.
            :param tick: The tick at which the candidate is added to the candidate store.
            :return: void
        """
        pass

    @abstractmethod
    def getCandidate(self, candidateHash: str) -> Tuple[ndarray, Set[int]]:
        """
            This method returns a candidate prototype given its hash along with the indices of the clusters that the
            sequence is a candidate for.

            :param candidateHash: The hash of the candidate prototypes that needs to be returned along with the indices
            of the clusters that the sequence is a candidate for.
            :return: The candidate prototype given its hash along with the indices of the clusters that the
            sequence is a candidate for.
        """
        pass

    @abstractmethod
    def lastUpdateCandidate(self, candidateHash: str) -> Optional[int]:
        """
            This method returns the tick at which the candidate was last updated.

            :param candidateHash: The hash of the candidate for which the tick at which it was last updated needs to
            be returned.
            :return: An integer representing the tick at which the candidate was last updated.
        """
        pass

    @abstractmethod
    def removeFromCandidates(self, candidateHash: str) -> None:
        """
            This method removes a candidate, identified by its hash, from the candidate store.

            :param candidateHash: The hash of the candidate that needs to be removed from the candidate store.
            :return: void
        """
        pass



"""
    This module contains the interface for a heuristic function that calculates the value of a (candidate) prototype
    given the representativeness and weight of the prototype.
"""

from abc import ABC, abstractmethod


class IPrototypeValue(ABC):

    @abstractmethod
    def evaluate(self, representativeness: float, weight: float) -> float:
        """
            This method evaluates the function that returns the linear combination of the representativeness and
            weight of some prototype.

            :param representativeness: The representativeness of some prototype.
            :param weight: The weight of some prototype.
            :return: The value of the prototype.
        """
        pass



"""
    This package contains implementations of evaluators that can assess the performance of the variants of the 'SeqClu'
    algorithm.
"""

from .basic_baseline_prototypes import BasicBaselinePrototypes
from .basic_evaluator import BasicEvaluator


__all__ = ["BasicBaselinePrototypes",
           "BasicEvaluator"]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="seqclupv",
    version="1.0.0",
    author="R. E. C. te Wierik",
    author_email="r.e.c.tewierik@student.tudelft.nl",
    description="An extension of the original 'SeqClu' algorithm that is characterized "
                "by voting for cluster prototypes.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rtewierik/seqclupv",
    project_urls={
        "Bug Tracker": "https://github.com/rtewierik/seqclupv/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    package_data={'': ['data/*/*.ts', 'data/handwriting/*']},
    install_requires=['fastdtw>=0.3.4',
                      'numpy>=1.20.3',
                      'seaborn>=0.11.1',
                      'matplotlib>=3.4.2',
                      'sklearn>=0.0',
                      'scipy>=1.6.3',
                      'pandas>=1.2.4',
                      'ipython>=7.24.1',
                      'xxhash>=2.0.2',
                      'psutil>=5.8.0',
                      'sktime>=0.6.1']
)



"""
    This package contains all implementations of data sources that are used by the 'SeqClu' algorithm to request
    data at every tick.
"""
from .fake_data_source import FakeDataSource


__all__ = ["FakeDataSource"]



"""
	This module contains a method that checks if some hash is part of a key that is a tuple of two hashes.
"""

from typing import Tuple


def isHashInKey(key: Tuple[str, str], sequenceHash: str) -> bool:
	"""
		This method checks if some hash is part of a key that is a tuple of two hashes.

		:param key: A key that is a tuple of two hashes.
		:param sequenceHash: The hash that needs to be compared to the key.
		:return: A boolean value indicating whether or not the incoming hash is part of the key.
	"""
	hashOne, hashTwo = key
	return hashOne == sequenceHash or hashTwo == sequenceHash



"""
    This module contains an interface for classes that are responsible for generating data that is passed to the
    'SeqClu' algorithm.
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Union, Optional

from numpy import ndarray


class IDataGenerator(ABC):

    # PROPERTIES #

    @property
    def classes(self) -> Optional[chr]:
        """
            This property stores the identifiers of items that are present in the data set.

            :return: A list of integers that represent the identifiers of
            the classes of the items that are present in the data set.
        """
        return self._classes

    @property
    def classDictionary(self) -> Dict[str, int]:
        """
            This property stores a mapping from class identifiers to indices.

            :return: A dictionary in which the keys are the identifiers of the classes of the items
            that are present in the data set and the values are the indices of the classes that can be used to
            retrieve the data that belong to a given class from a two-dimensional list containing all items in the data
            set.
        """
        return self._classDictionary

    @property
    def data(self) -> Optional[List[ndarray]]:
        """
            This property stores the data in the data set.

            :return: A list of sequences that are present in the data set.
        """
        return self._data

    @property
    def distances(self) -> Optional[ndarray]:
        """
            This property stores the pair-wise distances of items in the data set.

            :return: A two-dimensional array of pair-wise distances where indices i,j represent the indices of the
            two items in the list of sequences stored in the 'data' property.
        """
        return self._distances

    @property
    def indices(self) -> Optional[List[int]]:
        """
            This property stores the indices of items in the data set.

            :return: A list of indices of items in the data set.
        """
        return self._indices

    @property
    def labels(self) -> Optional[List[chr]]:
        """
            This property stores the labels of items in the data set.

            :return: A list of labels of items in the data set.
        """
        return self._labels

    @property
    def numPrototypes(self) -> int:
        """
            This property stores the number of prototypes that should be drawn from the data set.

            :return: An integer representing the number of prototypes that should be drawn from the data set.
        """
        return self._numPrototypes

    @abstractmethod
    def generateData(self) -> List[Union[ndarray, list]]:
        """
            This method generates a data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.

            :return: A data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.
        """
        pass

    def __init__(self, numPrototypes: int) -> None:
        """
            This method initializes the properties of the class.

            :param numPrototypes: The number of prototypes that will be used in the 'SeqClu' algorithm.
            :return: void
        """
        self._classes = None
        self._classDictionary = None
        self._data = None
        self._distances = None
        self._indices = None
        self._labels = None
        self._numPrototypes = numPrototypes



"""
	This module contains a method that hashes an incoming sequence using the Python package 'xxhash'.
"""

from numpy import ndarray
import xxhash


def hashSequence(sequence: ndarray) -> str:
	"""
		This method hashes an incoming sequence using the Python package 'xxhash'.

		:param sequence: The incoming sequence that needs to be hashed.
		:return: The hash of the incoming sequence.
	"""
	return xxhash.xxh32(sequence).hexdigest()
